// ESA (C) 2000-2018 
//
// This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
//
//    SAS is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    SAS is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
//
// Efluxer.cc
//
// Authors:  Richard Saxton (VEGA/ESA)
//           ESA XMM SOC
//
// Description: Task to produce a fluxed spectrum
//
//  This is a direct translation of the IDL script deresp.pro
//  written by Pedro Rodriguez
//
//#include <ctime>

#include "CalServer.h"
#include "Dal.h"

#include "Efluxer.h"

#include "Tracer.h"
#include "errstr.h"
#include "param.h"
#include "slatec.h"

#include "RMF.h"

using namespace std;

/* Method: Efluxer() - Constructor */
Efluxer::Efluxer()
{
///  PROC;

  /* Get the parameters from this command line */
    string srcfile = stringParameter("spectrumset");
    string bgdfile = stringParameter("backgndset");
    string arffile = stringParameter("arfset");
    string rmffile = stringParameter("rmfset");
    string outfile = stringParameter("fluxedset");
    //double tol = realParameter("tol"); // RMF tolerance

  /* Read in response file */
    //RMF rmf;
    //Response *resp = rmf.fixrmf(rmffile);
    RMF *rmf = new RMF(rmffile);
    Response *resp = rmf->getRSP();

  /* Open the spectrum and get the parameters */
    DataSet * spectrumset = dataSetServer->open( srcfile, DataSet::Read );
    Table * spectrumtable = spectrumset->table("SPECTRUM");

   // Get the counts
    Column *cntscol = spectrumtable->column("COUNTS");
    ColumnData *cntsdata = cntscol->data();
    int32 *scounts = cntsdata->int32Data();

   // Read a couple of keywords
    double texp=spectrumtable->attribute("EXPOSURE")->asReal();
    double sbackscal=spectrumtable->attribute("BACKSCAL")->asReal();
    //unsigned npi=spectrumtable->attribute("NAXIS2")->asInt();  // No. of chans
    unsigned npi=spectrumtable->rows();

   // Set up the background subtracted counts array
    vector<double> bsubcnts(npi);
    for (unsigned i=0; i<npi; ++i)
	bsubcnts[i]=scounts[i]*1.0;

  /* Subtract the background counts if file has been given */
    if (booleanParameter("withbgdset")) {
       DataSet * backgndset = dataSetServer->open( bgdfile, DataSet::Read );
       Table * backgndtable = backgndset->table("SPECTRUM");
       
       Column *bcntscol = backgndtable->column("COUNTS");
       ColumnData *bcntsdata = bcntscol->data();
       int32 *bcounts = bcntsdata->int32Data();

       double bbackscal=backgndtable->attribute("BACKSCAL")->asReal();

      // Subtract the background counts correcting for areas
       for (unsigned i=0; i<npi; ++i)
              bsubcnts[i] = max( 0.0, 
                    (bsubcnts[i] - bcounts[i] * sbackscal/bbackscal) );

      // Clean
       delete bcntsdata;
       dataSetServer->close(backgndset);
    }

  // Clean up
    delete cntsdata;

  // Calculate the PI channel centres and widths
    if (npi != resp->ebhi.size())
       errstr<<"Inconsistent number of channels in the RMF "
               "and spectral files"<<Sas::error("BadEnergyNum");

    vector<double> pi(npi), dpi(npi);

    for (unsigned i=0; i<npi; ++i)
       {
          pi[i] = (resp->ebhi[i] + resp->eblo[i]) / 2.0;
          dpi[i] = (resp->ebhi[i] - resp->eblo[i]);
       }

//    npi = pi.size(); // Note reassigning of npi (should be same as before)

  /* Open and read the ARF */
    DataSet * arfset = dataSetServer->open( arffile, DataSet::Read );
    Table * arftable = arfset->table("SPECRESP");

       
    Column *enlcol = arftable->column("ENERG_LO"),
           *enhcol = arftable->column("ENERG_HI"),
           *arfcol = arftable->column("SPECRESP");
    ColumnData *enldata = enlcol->data(),
	       *enhdata = enhcol->data(),
	       *arfdata = arfcol->data();

    real32 *elo_arf = enldata->real32Data(),
	    *ehi_arf = enhdata->real32Data(), 
	    *arf = arfdata->real32Data();

    //unsigned nenerg = arfdata->elements();  // No. of energies in ARF and RMF
    unsigned nenerg = arftable->rows();  // No. of energies in ARF and RMF

    vector<double> earf(nenerg);
    vector<double> barf(nenerg);

    float zero = 0.0;

    for (unsigned i=0; i<nenerg; ++i)
       {
         // Set centre energy values and widths
          earf[i] = (ehi_arf[i] + elo_arf[i]) / 2.0;
          barf[i] = (ehi_arf[i] - elo_arf[i]);

         /* Make sure that ARF values aren't NAN's - from IDL code */
          arf[i] = max(zero,arf[i]);
       }

   // Find the peak value and position of the RMF at each energy
    vector<double> rmax(nenerg);
    vector<int> imax(nenerg);
    double rmfmin=9999999.0;

    for (unsigned e=0; e<nenerg; ++e) { // Loop over the energies
       rmax[e]=0.0;
       for (unsigned i=0; i<npi; ++i) { // Loop over the PIs
         // Find the peak RMF value at this energy
          if ( resp->matrix[e][i] > rmax[e] ) {
            rmax[e] = resp->matrix[e][i];
            imax[e] = i;
          }
       }

      // Set the minimum non-zero peak value for all energies
       if (rmax[e] > 0.0) rmfmin = min(rmax[e], rmfmin);

    } // End of loop over energies

   // Make any peaks at value 0 = minimum non-zero peak value
    for (unsigned e=0; e<nenerg; ++e) 
        if (rmax[e] < rmfmin) rmax[e]=rmfmin;

   /* Make an array of Gaussian widths */
    vector<double> gwidth_orig(nenerg+14);
    vector<double> gwidth(nenerg);
    for (unsigned e=0; e<nenerg; ++e) {
        gwidth[e] = 1.0 / rmax[e] / sqrt(2.0*M_PI);
        gwidth_orig[e] = gwidth[e];
    }

   /* Smooth these widths by averaging over 15 energies. NB this
      doesn't handle the edges */
    for (unsigned e=7; e<nenerg-7; ++e) {
        double widout = 0.0;
        for (unsigned ee=e-7; ee<=e+7; ++ee) 
            widout +=gwidth_orig[ee];

        gwidth[e] = widout/15.0; 
    }

   // Make a minbin array
    vector<double> minbin(nenerg);
    for (unsigned e=0; e<nenerg; ++e) {
        minbin[e] = dpi[imax[e]] *  max(gwidth[e], 2.0); }

   /* Make an array of indices */
    vector<int> ii;
    int ix=nenerg-1;
    ii.push_back(ix);
    double rlo=0.0;

    while (ix > 0) {

      double testenl = ehi_arf[ix] - minbin[ix];

     /* Find the index of the ARF low_energy bounds containing testenl */
      if ( testenl < elo_arf[0] )
         rlo=0.0;
      else if ( testenl > elo_arf[nenerg-1] )
         rlo=double(nenerg)-1.0;
      else {
        for (unsigned lp=0; lp<(nenerg-2); ++lp) {

          if ( testenl >= elo_arf[lp] && 
                    testenl <=  elo_arf[lp+1]) {

             rlo = lp + (testenl - elo_arf[lp]) / 
                            (elo_arf[lp+1] - elo_arf[lp]);
          }  
        }
      }

      int jj = INT(floor(rlo));

      ii.push_back(jj);  // Add this index to the vector

      ix = jj-1; // Change the loop argument
    }

   // Reverse the index array
    reverse(ii.begin(), ii.end());

    unsigned nii = ii.size() - 1; 

   // Set up a couple of arrays
    vector<int> is;
    vector<int> ie;

    for (unsigned e=0; e<nii; ++e) {
        is.push_back(ii[e]);
        ie.push_back(ii[e+1]-1);
    }

   // Some energy arrays
    vector<double> energc;
    vector<double> energb;

    for (unsigned e=0; e<nii; ++e) {
        energc.push_back( (ehi_arf[ie[e]] + elo_arf[is[e]]) /2.);
        energb.push_back( (ehi_arf[ie[e]] - elo_arf[is[e]]) );
    }

   // Some more arrays
    //vector<vector<double> > rmat(npi);
    vector<vector<double> > rmat;
    vector<double> specresp(nii);
    vector<unsigned> k;
    vector<double> tempresp(nii);

   // Find where ie-is is positive
    for (unsigned e=0; e<nii; ++e) 
        if ( (ie[e] - is[e]) > 0 ) k.push_back(e);

    for (unsigned e=0; e<npi; ++e) {

     unsigned nk=k.size();
     for (unsigned j=0; j<nk; ++j) {
        unsigned i=k[j];

        specresp[i]=0.0;

       // Sum the spec response elements
        double totee=0.0;
        for (int ee=is[i]; ee<=ie[i]; ++ee) {
            specresp[i] += arf[ee] * barf[ee] / energb[i];
            totee +=barf[ee];
        }

       // Loop over each energy and average the response for this chan
           double totresp = 0.0;

           for (int ee=is[i]; ee<=ie[i]; ++ee) 
               totresp += resp->matrix[ee][e] * barf[ee];

           // Normalise the response if total is non-zero
            if (totee > 0.0)
               tempresp[i] = totresp / double(totee);
            else
               tempresp[i] = 0.0;


     } // End of loop over j = PI centres
     rmat.push_back(tempresp);
    } // End of loop over energies

   // Set up counts and error arrays
    vector<double> cts(nii), ects(nii);

   // Find the nearest (lower) index containing the right energy
    vector<int> ilo, ihi;
    double rhi=0.0;
    double testenl=0.0, testenh=0.0;
    int ihisave=0;

    for (unsigned e=0; e<nii; ++e) {

      testenl = energc[e]-energb[e]/2.;
      testenh = energc[e]+energb[e]/2.;

      for (unsigned i=0; i<npi-1; ++i) {
        //if ( testenl >= (pi[i]-dpi[i]/2.) && 
        if ( testenl > (pi[i]-dpi[i]/2.) && 
                    testenl <= (pi[i+1] - dpi[i+1]/2.) ) {

           rlo = i + (testenl - (pi[i] - dpi[i]/2.)) / 
                            (pi[i+1] - pi[i]);
           //rlo = double(i);
        }
        if ( testenh >= (pi[i]+dpi[i]/2.) && 
                    testenh < (pi[i+1] + dpi[i+1]/2.) ) {
                    //testenh <= (pi[i+1] + dpi[i+1]/2.) ) {

           rhi = i + (testenh - (pi[i] + dpi[i]/2.)) / 
                                         (pi[i+1] - pi[i]);
        }
      }

     // New plan to make ilo[e] = ihi[e-1]+1
      if (e > 0 ) 
         ilo.push_back(ihisave+1);
      else {
         if (testenl < pi[0]) 
            ilo.push_back(0);
         else if (testenl > pi[npi-1])
            ilo.push_back(npi-1);
         else
            ilo.push_back(INT(floor(rlo)));
      }

     // Round up to make sure that ihi>ilo
      if (testenh < pi[0]) 
         ihi.push_back(0);
      else if (testenh > pi[npi-1])
         ihi.push_back(npi-1);
      else {
         ihi.push_back(INT(ceil(rhi)));
         ihisave=INT(ceil(rhi));
      }
    }

   // Find the cumulative spectrum
    vector<double> spec_cum(npi);
    spec_cum[0]=bsubcnts[0];
    double maxscum=0.0;

    for (unsigned i=1; i<npi; ++i) {
        spec_cum[i] = spec_cum[i-1] + bsubcnts[i];
        maxscum = max(maxscum,spec_cum[i]);
    }

   // Make array of places where ihi > ilo
    vector<int> g;
    for (unsigned i=0; i<ilo.size(); ++i) {
        if ( ( ihi[i] - ilo[i] ) > 0 ) g.push_back(i);
    }
    int ng = g.size();

   // Find the first time that something happens
    int i2=ng-1;  // Set default
    for (int lp=0; lp<ng; ++lp) {

        if (spec_cum[ilo[g[lp]]] == maxscum) {
           i2 = lp;
           break;
        }
    }

    int i = g[i2];  // Index for coming equations

   // Find the final index
    int iend=0;
    for (int lp=(ng-1); lp>=0; --lp) {
        if (spec_cum[ilo[g[lp]]] == 0 && iend==0) {
           iend = max((lp-1),0);
        }
    }

   // Initialise the bscale array
    vector<double> bscale(npi);

    double topen = energc[i]+energb[i]/2.;
    double boten = energc[i]-energb[i]/2.;

    for (int lp=ilo[i]; lp<=ihi[i]; ++lp) { 
       // bscale[lp] = (min ( (pi[lp] + dpi[lp]/2.0 ), topen ) -
         //                max( (pi[lp] - dpi[lp]/2.0), boten)) / dpi[lp];
      //  bscale[lp] = max( min(bscale[lp], 1.0), 0.0);

       bscale[lp]=1.0;
    }

   // Make the cts array
    cts[i] = 0.0;
    ects[i]= 0.0;
    double widsum=0.0;
    double cntsum=0.0;

   // Set up the counts matrix and set the first element
    vector<double> cmat(nii);

    for (unsigned e=ilo[i]; e<npi; ++e)  
                  cmat[i] += rmat[e][i];

    // Sum of counts between ilo and ihi
    for (int lp=ilo[i]; lp<=ihi[i]; ++lp) {
        cntsum += bsubcnts[lp];
        widsum += dpi[lp];
    }
    // Divide by the sum of the widths
    cts[i] = cntsum / cmat[i] / widsum;
    ects[i] = sqrt(max(cntsum,1.0)) / cmat[i] / widsum;

   // Normalise by specresp[i] if non-zero
    if (specresp[i] > 0.0) {
       cts[i] = cts[i] / specresp[i];
       ects[i] = ects[i] / specresp[i];
    }

   // Initialise the "fudge" arrays
    vector<double> fudge(npi);
    vector<double> efudge(npi);

   /* Initialise the fudge, efudge arrays */
    for (unsigned lp=0; lp<npi; ++lp) {  // Loop over channels

      // Sum the RMF contributions at this channel
       fudge[lp] = 0.0;
       efudge[lp] = 0.0;

       for (unsigned e=0; e<nii; ++e)  {
         fudge[lp] += rmat[lp][e] * cts[e] * specresp[e] * energb[e];
         efudge[lp] += rmat[lp][e] * ects[e] * specresp[e] * energb[e];
       }
    }
   // Set a couple of constants
    //int jsta = i2-2;
    int jsta = i2-1;
    double cscale =1.0;

   // Loop over channel elements
    for (int j=jsta; j>=iend; --j) {

        //unsigned i = g[j];
        i = g[j];  // Probably the same 'i' as above

      /* Calculate a scaling factor for each channel in the
         right range */
        topen = energc[i]+energb[i]/2.;
        boten = energc[i]-energb[i]/2.;

        for (int lp=ilo[i]; lp<=ihi[i]; ++lp) { 

            //bscale[lp] = (min ( (pi[lp] + dpi[lp]/2.0 ), topen ) -
              //           max( (pi[lp] - dpi[lp]/2.0), boten)) / dpi[lp];
            //bscale[lp] = max( min(bscale[lp], 1.0), 0.0);

           bscale[lp]=1.0;
        }

      /* Calculate the fudge factors ??? */
       for (unsigned lp=0; lp<npi; ++lp) {  // Loop over channels

         // Sum the RMF contributions at this channel
          fudge[lp] = 0.0;
          efudge[lp] = 0.0;

          for (unsigned e=g[j+1]; e<nii; ++e)  {
             fudge[lp] += rmat[lp][e] * cts[e] * specresp[e] * energb[e];
             efudge[lp] += rmat[lp][e] * ects[e] * specresp[e] * energb[e];
          }
       }

      // Loop over affected PI elements
       for (int lp=i; lp<(g[j+1]); ++lp) { 

         // Sum up the matrix elements
          double rsum=0.0, fsum=0.0, dpisum=0.0, csum=0.0, esum=0.0;

          int mine = max( (ilo[i]), 0);
          for (unsigned e=mine; e<npi; ++e)  
                           rsum += rmat[e][lp];

          cmat[lp] = rsum / (g[j+1]-i);

         // Calculate the counts array
          for (int e=ilo[i]; e<=ihi[i]; ++e) {
	      csum += bsubcnts[e];
              fsum += fudge[e];
              dpisum += dpi[e] * bscale[e];
              esum += efudge[e]*efudge[e];
          }

          cts[lp] = max(csum - cscale * fsum, 0.0) / cmat[i] / dpisum;
          ects[lp] = sqrt(csum + max(cscale * esum, 1.0)) / dpisum
                           /cmat[i];

         // Normalise by specresp if non-zero
          if (specresp[lp] !=0.0) {
             cts[lp] /= specresp[lp];
             ects[lp] /= specresp[lp];
          }
       } // End of lp loop
    } // End of j loop 

   /* Write out the final file */
   // Start by copying the input spectrum
    DataSet * outputset = dataSetServer->open( outfile,
                                      DataSet::Create);

    Table * outtab = outputset->addTable("FluxedSpectrum", nii);

   // Make energy columns
    Column * energyCol  = outtab->addColumn("ENERGY",
                            Column::Real32, "Energy", "keV");
    ColumnData *engydata = energyCol->data();
    real32 *energy = engydata->real32Data();
    Column * ewidthCol  = outtab->addColumn("ENERGY_BIN",
                            Column::Real32, "Energy bin", "keV");
    ColumnData *ewiddata = ewidthCol->data();
    real32 *ewidths = ewiddata->real32Data();

   // Make a flux column
    Column *ofluxCol = outtab->addColumn("FLUX",
                            Column::Real32, "flux", "ergs/s/cm2/keV");
    ColumnData *ofdata = ofluxCol->data();
    real32 *ospec = ofdata->real32Data();

   // Make an error column
    Column * ecntsCol  = outtab->addColumn("ERROR",
                            Column::Real32, "flux error", "ergs/s/cm2/keV");
    ColumnData *ecntsdata = ecntsCol->data();
    real32 *espec = ecntsdata->real32Data();

   // Write flux into the "counts" column
    double echarge=1.6021765e-09; // Electron charge
    for (unsigned lp=0; lp<nii; ++lp) {
      ospec[lp] = cts[lp] * echarge * energc[lp] /  max(texp,1.0);
      espec[lp] = ects[lp] * echarge * energc[lp] /  max(texp,1.0);
      energy[lp] = energc[lp];
      ewidths[lp] = energb[lp]/2.0;
    }

   // Transfer header attributes
    rmf->writeAttributes(spectrumset, outtab);

   // Close down and go home
    delete enldata;
    delete enhdata;
    delete arfdata;  // Clean up

    delete ofdata;
    delete ecntsdata;
    delete engydata;
    delete ewiddata;

    dataSetServer->close(spectrumset);
    dataSetServer->close(arfset);
    dataSetServer->close(outputset);
    delete rmf;

//ENDPROC;
}


/* Destructor */
Efluxer::~Efluxer()
{
  PROC;

  ENDPROC;
}



